#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include "protocol.h"
#define STAT_WAITING 1<<0
#define STAT_READING 1<<1
#define STAT_WRITING 1<<2

int add_byte(char c);
void reset_packet();

static struct termios oldtio, newtio, contio;
static int portfd = -1;
static struct packet_struct {
  unsigned char data[15];
  unsigned char len;
  unsigned char sum;
  unsigned char wrote;
} packet;

static struct incoming_packet {
  unsigned char data[15];
  unsigned char len;
  unsigned char sum;
  unsigned char recv;
} ipacket;

/* switch baud rate */
void serial_baud(speed_t speed)
{
        assert(portfd >= 0);
        tcdrain(portfd);
        usleep(50 * 1000);      /* 50 ms sleep; arbitrary */
        cfsetispeed(&newtio, speed);
        cfsetospeed(&newtio, speed);
        tcsetattr(portfd, TCSANOW, &newtio);
}

/* close serial port and restore settings */
void serial_close(void)
{
        assert(portfd >= 0);
        tcsetattr(portfd, TCSANOW, &oldtio);
        if (close(portfd) < 0)
          exit(1);
        portfd = -1;
}

/* write or die */
ssize_t xwrite(int fd, const void *buf, size_t count)
{
        ssize_t retval;
        do {
                retval = write(fd, buf, count);
        } while ((retval < 0) && (errno == EINTR));
        if (retval < 0)
                exit(1);
        return retval;
}


/* send a character on the serial port */
void put_char(unsigned char c)
{                       
        assert(portfd >= 0);
        xwrite(portfd, &c, 1);
}

/* read or die */
ssize_t xread(int fd, void *buf, size_t count)
{
        ssize_t retval;
        do {
                retval = read(fd, buf, count);
        } while ((retval < 0) && (errno == EINTR));
        if (retval < 0)
                exit(1);
        return retval;
}

/* wait for a character on the serial port */
unsigned char get_char(void)
{
        unsigned char c;
        assert(portfd >= 0);
        xread(portfd, &c, 1);
        printf("read %2x\n",c);
        return c;
}

int get_char_timeout(int msecs)
{       
  fd_set fds;   
        struct timeval tv;
                
        assert(portfd >= 0);
                
        FD_ZERO(&fds);
        FD_SET(portfd, &fds);
        tv.tv_sec = 0;
        tv.tv_usec = msecs * 1000;
                
        select(portfd+1, &fds, NULL, NULL, &tv);
        if (FD_ISSET(portfd, &fds))
                return get_char();
        return -1;  
}                   

void serial_open(const char *dev)
{
        portfd = open(dev, O_RDWR | O_NOCTTY);
        if (portfd < 0)
                exit(1);

        /* save current port settings */
        if (tcgetattr(portfd, &oldtio) < 0)
                exit(1);

        /* configure new port settings: 9600 8N1 */
        memset(&newtio, 0, sizeof(newtio));
        newtio.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
        newtio.c_iflag = IGNPAR;
        newtio.c_oflag = 0;
        /* set input mode (non-canonical, no echo,...) */
        newtio.c_lflag = 0;
        newtio.c_cc[VTIME] = 0; /* inter-character timer unused */
        newtio.c_cc[VMIN] = 1;  /* blocking read until 1 char received */

        /* install new port settings */
        tcflush(portfd, TCIFLUSH);
        if (tcsetattr(portfd, TCSANOW, &newtio) < 0)
                exit(1);
}

int write_packet(int len, char str[])
{
  int i;
printf("in write_packet\n");
  reset_packet();
  for(i=0;i<len;i++) {
    add_byte(str[i]);
  }
  return write_byte();
}

int add_byte(char c)
{
  if (packet.len < 15 && !packet.wrote) {
    packet.data[packet.len++] = c;
    packet.sum = (packet.sum - c)&0xff;
  } else {
    return 1;
  }
  return 0;
}

void reset_packet()
{
  packet.wrote=packet.len=packet.sum=0;
}

int write_byte()
{
  if (packet.wrote < packet.len) {
    printf("Writing 0x%02x\n", packet.data[packet.wrote]);
    put_char(packet.data[packet.wrote++]);
    return STAT_WRITING;
  } else if (packet.wrote == packet.len) {
    printf("Writing 0x%02x\n", packet.sum);
    put_char(packet.sum);
    packet.wrote++;
    return STAT_WRITING;
  } else {
    printf("error writing packet\n");
    return STAT_WAITING;
  }
}

int process_ipacket()
{
  switch(ipacket.data[0]) {
    case PING:
      reset_packet();
      add_byte(PONG);
      write_byte();
      return STAT_WRITING;
    case SESINFO:
      printf("got 8051 initialized in mode %02x, run type %02x, lot %02x %02x\n", ipacket.data[1], ipacket.data[2], ipacket.data[3], ipacket.data[4]);
      return STAT_WAITING;
    case REQ_TOT_TRK: /* Always return 10 tracks for now */
      reset_packet();
      add_byte(TRACKS);
      add_byte(0x00);
      add_byte(0x0A);
      add_byte(ipacket.data[1]);
      add_byte(ipacket.data[2]);
      write_byte();
      return STAT_WRITING;
    case Q_TOT_DSK: /* Always return 10 disks for now */
      reset_packet();
      add_byte(DISKS);
      add_byte(0x00);
      add_byte(0x0A);
      write_byte();
      return STAT_WRITING;
    case REQ_DISK_NAME: /* Always return "Emulation" */
      reset_packet();
      add_byte(0x50);
      add_byte(0x00);
      add_byte(0x00);
      add_byte('E');
      add_byte('m');
      add_byte('u');
      add_byte('l');
      add_byte('a');
      add_byte('t');
      add_byte('i');
      add_byte('o');
      add_byte('n');
      add_byte(0x00);
      add_byte(0x00);
      add_byte(0x00);
      write_byte();
      return STAT_WRITING;
    case REQ_TRACK_NAME: /* Always return "Trackname" */
      reset_packet();
      add_byte(0x50);
      add_byte(0x01);
      add_byte(0x00);
      add_byte('T');
      add_byte('r');
      add_byte('a');
      add_byte('c');
      add_byte('k');
      add_byte('n');
      add_byte('a');
      add_byte('m');
      add_byte('e');
      add_byte(0x00);
      add_byte(0x00);
      add_byte(0x00);
      write_byte();
      return STAT_WRITING;
    case RAND_FORWARD: /* Always return next track as Disk 7 Track 9 */
      reset_packet();
      add_byte(PLAYING_TRACK);
      add_byte(0x00);
      add_byte(0x09);
      add_byte(0x00);
      add_byte(0x07);
      write_byte();
      return STAT_WRITING;
    case REQ_SETTINGS:
      reset_packet();
      add_byte(CUR_SETTINGS);
      add_byte(0x19);
      add_byte(0x00);
      add_byte(0x00);
      add_byte(0x00);
      write_byte();
      return STAT_WRITING;
    case PLAY:
    case STOP:
    case UCID:
    case SHUTDOWN:
      return STAT_WAITING;
  }
}

int recv_byte(char c)
{
  printf("In recv_byte\n");
  if (ipacket.recv < ipacket.len) {
    ipacket.data[ipacket.recv++] = c;
    ipacket.sum = (ipacket.sum - c)&0xff;
    printf("packet %02x added to stack\n", c);
    return 0;
  } else if (ipacket.len == ipacket.recv) {
    if (c == ipacket.sum) {
      put_char(PACKET_ACK);
      printf("got good packet\n");
      return 1;
    } else {
      printf("Invalid checksum\n");
      return -1;
    }
  } else {
    printf("I Should never have gotten here (len=%d, recv=%d)!\n", ipacket.len, ipacket.recv);
    return -1;
  }
}


void reset_ipacket()
{
  printf("RESET IPACKET\n");
  ipacket.recv=ipacket.len=ipacket.sum=0;
}

int init_ipacket(char c)
{
  switch(c) {
    case PING:
      printf("Got PING begin\n", c);
      reset_ipacket();
      ipacket.len=1;
      recv_byte(c);
      return 1;
    case SESINFO:
      printf("Got 8051 Session Info header\n");
      reset_ipacket();
      ipacket.len=5;
      recv_byte(c);
      return 1;
    case UCID:
      printf("Got UCID header\n");
      reset_ipacket();
      ipacket.len=7;
      recv_byte(c);
      return 1;
    case REQ_TOT_TRK:
      printf("Got QUERY NUM TRACKS begin\n", c);
      reset_ipacket();
      ipacket.len=3;
      recv_byte(c);
      return 1;
    case Q_TOT_DSK:
      printf("Got QUERY NUM DISKS begin\n", c);
      reset_ipacket();
      ipacket.len=1;
      recv_byte(c);
      return 1;
    case PLAY:
      printf("Got PLAY begin\n", c);
      reset_ipacket();
      ipacket.len=7;
      recv_byte(c);
      return 1;
    case REQ_SETTINGS:
      printf("Got REQ_SETTINGS begin\n");
      reset_ipacket();
      ipacket.len=1;
      recv_byte(c);
      return 1;
    case REQ_DISK_NAME:
      printf("Got QUERY DISK NAME begin\n", c);
      reset_ipacket();
      ipacket.len=3;
      recv_byte(c);
      return 1;
    case REQ_TRACK_NAME:
      printf("Got QUERY TRACK NAME begin\n", c);
      reset_ipacket();
      ipacket.len=5;
      recv_byte(c);
      return 1;
    case REPEAT_TRK:
      printf("Got REPEAT TRACK begin\n", c);
      reset_ipacket();
      ipacket.len=1;
      recv_byte(c);
      return 1;
    case REPEAT_DSK:
      printf("Got REPEAT DISK begin\n", c);
      reset_ipacket();
      ipacket.len=1;
      recv_byte(c);
      return 1;
    case REPEAT_OFF:
      printf("Got REPEAT OFF begin\n", c);
      reset_ipacket();
      ipacket.len=1;
      recv_byte(c);
      return 1;
    case RAND_ON:
      printf("Got RANDOM DISK begin\n", c);
      reset_ipacket();
      ipacket.len=6;
      recv_byte(c);
      return 1;
    case MAG_RAND_ON:
      printf("Got RANDOM MAG begin\n", c);
      reset_ipacket();
      ipacket.len=6;
      recv_byte(c);
      return 1;
    case RAND_OFF:
      printf("Got RANDOM OFF begin\n", c);
      reset_ipacket();
      ipacket.len=1;
      recv_byte(c);
      return 1;
    case RAND_FORWARD:
      printf("Got RANDOM_NEXT begin\n", c);
      reset_ipacket();
      ipacket.len=1;
      recv_byte(c);
      return 1;
    case STOP:
      printf("Got STOP begin\n", c);
      reset_ipacket();
      ipacket.len=1;
      recv_byte(c);
      return 1;
    case SHUTDOWN:
      printf("Got SHUTDOWN begin\n", c);
      reset_ipacket();
      ipacket.len=1;
      recv_byte(c);
      return 1;
    default:
      printf("Unknown byte 0x%02x\n", c);
      return 0;
  }
}

int main()
{
  fd_set fds;
  struct timeval tv;
  int retval;
  unsigned char c;
  char packet[16] = {0x10, 0xc, 0x2, 0xff, 0x0, 0x14, 0x5};
  int status=STAT_WAITING;

  tv.tv_sec=0;
  tv.tv_sec=100000;
  serial_open("/dev/ttyS1");
  serial_baud(B115200);

  printf("Emulating 51d\n");
  put_char(0xf9); /* Test phat proto version - only know V2 */
  if (get_char() != 0x53) {
    printf("Unexpected response, help\n");
    exit(1);
  }

  status=write_packet(7, packet);

  while(1) {
    FD_ZERO(&fds);
    FD_SET(portfd, &fds);
    retval = select(portfd+1, &fds, NULL, NULL, &tv);
    if (retval > 0 && FD_ISSET(portfd, &fds)) {
      if (1 == read(portfd, &c, 1)) {
        printf("read 0x%02x\n",c);
        switch(status) {
          case STAT_WAITING:
            printf("STAT_WAITING\n");
            if (init_ipacket(c) == 1)
              status=STAT_READING;
            break;
          case STAT_READING:
            printf("STAT_READING\n");
            retval = recv_byte(c);
            switch (retval) {
              case -1:
                reset_ipacket();
                printf("new STAT_WAITING\n");
                status=STAT_WAITING;
                break;
              case 0:
                break;
              case 1:
                status=process_ipacket();
                break;
            }
            break;
          case STAT_WRITING:
            printf("STAT_WRITING\n");
            switch(c) {
              case PACKET_ACK:
                printf("new STAT_WAITING\n");
                reset_packet;
                status=STAT_WAITING;
                break;
              case HEADER_ACK:
              case BLOCK_ACK:
                status = write_byte();
                break;
              default:
                status=STAT_WAITING;
                if (init_ipacket(c) == 1)
                  status=STAT_READING;
                break;
            }
            break;
        }
      }
    }
  }
}
